@@ -30,6 +30,13 @@ width = 20
 status_box_height =6
 tilesize = 32
 
+#used for yes/no dialog boxes.
+active_button = 0
+return_from_yesno = StringVar()
+
+#used to prevent drawing problems.
+global already_refreshed
+already_refreshed = 0
 
 #max number of messages to keep.
 max_messages = 20
@@ -66,6 +73,7 @@ def print_message(message):
 
 #save game
 def save_game(event=0):
+	if action.has_dialog == 1: return 0
 	g.savegame(g.name)
 	print_message("** Game Saved **");
 
@@ -79,11 +87,20 @@ def dead_yet():
 
 #back to main menu. Called on death.
 def close_window(event=0):
-	if g.hp < 1 or show_yesno("Leave this game?"):
-		try:
-			window_main.destroy()
-		except TclError:
-			pass
+	if action.has_dialog == 1: return
+	if canvas_map.winfo_exists():
+		if g.hp < 1 or show_yesno("Leave this game?"):
+			try:
+				window_main.destroy()
+			except TclError:
+				pass
+	else: #if called from inv.
+		if g.hp < 1 or inv.inv_show_yesno("Leave this game?"):
+			try:
+				window_main.destroy()
+			except TclError:
+				pass
+
 
 #Activate any scripting associated with a tile. X and Y are relative
 #to the location of the player.
@@ -111,19 +128,24 @@ def activate_scripting(x, y, z):
 def refreshmap():
 	#redraw the tiles
 	half_map = mapsize/2 #division is nasty. Only do it once.
-	canvas_map.delete("all")
+	canvas_map.delete(ALL)
 	for x in range(mapsize):
 		for y in range(mapsize):
 			for picture in findtile(x-half_map, y-half_map):
 				canvas_map.create_image(x*g.tilesize, y*g.tilesize,
-					anchor=NW, image=picture)
-
-
-	canvas_map.create_image(mapsize*g.tilesize/2, mapsize*g.tilesize/2,
-		image=g.tiles[g.cur_hero])
+					anchor=NW, image=picture,
+					tags=str(g.xgrid+x-half_map)+"."+str(g.ygrid+y-half_map))
 
+	refreshhero()
 	recalc_stats()
 
+#refreshs the hero; faster than refreshing the whole map.
+def refreshhero():
+	canvas_map.delete("hero")
+	canvas_map.create_image(mapsize*g.tilesize/2, mapsize*g.tilesize/2,
+		image=g.tiles[g.cur_hero], tags="hero")
+
+
 #called to process the onload portion of a level. Called whenever entering
 #a level. Set g.zgrid properly before calling.
 def process_onload():
@@ -210,7 +232,8 @@ def move_hero(x, y):
 		if window_main.state() != "normal" or action.has_dialog == 1:
 			return
 
-		#check for any objects of interest.
+		#check for any objects of interest. Called before moving, in case
+		#the scripting changes the walkable status.
 		activate_scripting(x, y, g.zgrid)
 
 		if g.iswalkable(x, y):
@@ -222,13 +245,76 @@ def move_hero(x, y):
 				if y ==  1: g.cur_hero = "people/hero_s.png"
 				if x == -1: g.cur_hero = "people/hero_w.png"
 				if x ==  1: g.cur_hero = "people/hero_e.png"
-				refreshmap()
+				#Sometimes, quick refresh is bad.
+				global already_refreshed
+				if already_refreshed == 1:
+					refreshhero()
+					already_refreshed = 0
+					return
+
+				#quick refresh.
+				#Moves the canvas over a bit
+				canvas_map.move(ALL, -x*tilesize, -y*tilesize)
+
+				#Refreshes the new part.
+				half_map = mapsize/2
+				refreshhero()
+				recalc_stats()
+				g.allow_move = 1
+				if y == -1:
+					for tilenum in range(mapsize):
+						if len(canvas_map.gettags(str(g.xgrid+
+							tilenum-half_map)+"."+str(g.ygrid-half_map))) == 1:
+							continue
+						canvas_map.delete(str(g.xgrid+
+							tilenum-half_map)+"."+str(g.ygrid-half_map))
+						for picture in findtile(tilenum-half_map, -half_map):
+							canvas_map.create_image(tilenum*g.tilesize, 0,
+							anchor=NW, image=picture, tags=str(g.xgrid+
+							tilenum-half_map)+"."+str(g.ygrid-half_map))
+				if y ==  1:
+					for tilenum in range(mapsize):
+						if len(str(g.xgrid+ tilenum-half_map)+"."+
+							str(g.ygrid+half_map)) == 1:
+							continue
+						canvas_map.delete(str(g.xgrid+ tilenum-half_map)+"."+
+							str(g.ygrid+half_map))
+						for picture in findtile(tilenum-half_map, half_map):
+							canvas_map.create_image(tilenum*g.tilesize,
+							g.tilesize*(mapsize-1), anchor=NW, image=picture,
+							tags=str(g.xgrid+ tilenum-half_map)+"."+
+							str(g.ygrid+half_map))
+				if x == -1:
+					for tilenum in range(mapsize):
+						if len(canvas_map.gettags(str(g.xgrid-half_map)
+							+"."+ str(g.ygrid+tilenum-half_map))) == 1:
+							continue
+						canvas_map.delete(str(g.xgrid
+							-half_map) +"."+ str(g.ygrid+tilenum-half_map))
+						for picture in findtile(-half_map, tilenum-half_map):
+							canvas_map.create_image(0, tilenum*g.tilesize,
+							anchor=NW, image=picture, tags=str(g.xgrid
+							-half_map) +"."+ str(g.ygrid+tilenum-half_map))
+				if x ==  1:
+					for tilenum in range(mapsize):
+						if len(str(g.xgrid+half_map) +"."+
+							str(g.ygrid+tilenum-half_map)) == 1:
+							continue
+						canvas_map.delete(str(g.xgrid+half_map) +"."+
+							str(g.ygrid+tilenum-half_map))
+						for picture in findtile(half_map, tilenum-half_map):
+							canvas_map.create_image(g.tilesize*(mapsize-1),
+							tilenum*g.tilesize, anchor=NW, image=picture,
+							tags=str(g.xgrid+half_map) +"."+
+							str(g.ygrid+tilenum-half_map))
+
 		else:
 			if y == -1: g.cur_hero = "people/hero_n.png"
 			if y ==  1: g.cur_hero = "people/hero_s.png"
 			if x == -1: g.cur_hero = "people/hero_w.png"
 			if x ==  1: g.cur_hero = "people/hero_e.png"
-			refreshmap()
+			refreshhero()
+			recalc_stats()
 		g.allow_move = 1
 	except TclError:
 		pass
@@ -271,8 +357,9 @@ def passturn():
 		if temp != -1:
 			start_battle(temp)
 			if dead_yet() != 1:
-				refreshmap()
-
+				#If I don't do this, the quick refresh wigs out.
+				global already_refreshed
+				already_refreshed = 1
 
 
 #Starts a battle. Takes the index of the monster in g.monster.monster
@@ -303,10 +390,84 @@ def show_dialog(line=None):
 	tkMessageBox.showinfo(g.game_name, line, parent=window_main)
 	return 1
 
-def show_yesno(line=None):
+#given a string, display that string, along with yes and no buttons.
+def show_yesno(line=""):
 	if window_main.state() != "normal":
 		return 0
-	return tkMessageBox.askyesno(g.game_name, line, parent=window_main)
+	action.has_dialog = 1
+	global active_button
+	active_button = 0
+	global return_from_yesno
+	return_from_yesno.set("0")
+	#text
+	canvas_map.create_text((mapsize*tilesize/2),
+		(6*mapsize*tilesize/7), text=line, anchor="s",
+		width=no_button.width()*4/3+yes_button.width()*4/3-6, tags="dialogtxt")
+	#generate the background size.
+	x1, y1, x2, y2 = canvas_map.bbox("dialogtxt")
+	if x1 > (mapsize*tilesize/2) - no_button.width():
+		x1 = (mapsize*tilesize/2) - no_button.width()
+	if x2 < (mapsize*tilesize/2) + yes_button.width():
+		x2 = (mapsize*tilesize/2) + yes_button.width()
+	#background
+	canvas_map.create_rectangle(x1-5, y1-5, x2+5, y2+no_button.height(),
+		fill="#5050ff", outline="Black", tags="background_box")
+	canvas_map.lift("dialogtxt")
+	#buttons
+	canvas_map.create_image((mapsize*tilesize/2),
+		(6*mapsize*tilesize/7), image=no_button_sel, anchor="ne", tags="no")
+	canvas_map.create_image((mapsize*tilesize/2),
+		(6*mapsize*tilesize/7), image=yes_button, anchor="nw", tags="yes")
+	#bindings
+	window_main.bind("<Up>", change_yesno_sel)
+	window_main.bind("<Right>", change_yesno_sel)
+	window_main.bind("<Left>", change_yesno_sel)
+	window_main.bind("<Down>", change_yesno_sel)
+	window_main.bind("<Return>", activate_yesno)
+	canvas_map.tag_bind("no", "<Enter>", change_yesno_to_no)
+	canvas_map.tag_bind("yes", "<Enter>", change_yesno_to_yes)
+	canvas_map.tag_bind("no", "<Button-1>", activate_yesno)
+	canvas_map.tag_bind("yes", "<Button-1>", activate_yesno)
+	#wait. Continue after activate_yesno() is run.
+	window_main.wait_variable(return_from_yesno)
+	#cleanup
+	try:
+		if window_main.state() != "normal": return 1
+	except TclError: return 1
+	bind_keys()
+	canvas_map.bind("<Button-1>", mouse_handler)
+	action.has_dialog = 0
+	canvas_map.delete("dialogtxt", "no", "yes", "background_box")
+
+	return active_button
+
+#when a yesno dialog box is active, switch to yes.
+def change_yesno_to_yes(event=None):
+	global active_button
+	active_button = 1
+	canvas_map.itemconfigure("no", image=no_button)
+	canvas_map.itemconfigure("yes", image=yes_button_sel)
+
+#when a yesno dialog box is active, switch to no.
+def change_yesno_to_no(event=None):
+	global active_button
+	active_button = 0
+	canvas_map.itemconfigure("no", image=no_button_sel)
+	canvas_map.itemconfigure("yes", image=yes_button)
+
+#when a yesno dialog box is active, switch the yes and no buttons.
+def change_yesno_sel(event=None):
+	if active_button == 0: change_yesno_to_yes()
+	else: change_yesno_to_no()
+
+#called on Return when the yesno dialog box is active.
+def activate_yesno(event=None):
+	global return_from_yesno
+	return_from_yesno.set("1")
+
+def cleanup(event=None):
+	activate_yesno()
+	window_main.destroy()
 
 #call to create window_main.
 def init_window_main(new_game = 0):
@@ -347,6 +508,18 @@ def init_window_main(new_game = 0):
 
 	name.set(g.name)
 
+	#yes/no buttons for dialogs.
+	global yes_button; global yes_button_sel
+	yes_button = ImageTk.PhotoImage(file=g.mod_directory + \
+		"/images/buttons/yes.png")
+	yes_button_sel = ImageTk.PhotoImage(file=g.mod_directory + \
+		"/images/buttons/yes_sel.png")
+	global no_button; global no_button_sel
+	no_button = ImageTk.PhotoImage(file=g.mod_directory + \
+		"/images/buttons/no.png")
+	no_button_sel = ImageTk.PhotoImage(file=g.mod_directory + \
+		"/images/buttons/no_sel.png")
+
 	#left (stats) side frame
 	global left
 	left = Frame(top, bg=bgcolour);
@@ -431,6 +604,7 @@ def init_window_main(new_game = 0):
 	free_move = 0
 
 	bind_keys()
+	window_main.protocol("WM_DELETE_WINDOW", cleanup)
 
 	#call newgame script
 	if new_game == 1:
